@@ -38,11 +38,12 @@ from . import Error, Type, Env, Expr, Tree, StdLib, Walker, _util
 
 def _find_doc(obj: Error.SourceNode):
     "find the containing document"
-    if hasattr(obj, "_doc4lint"):
-        return getattr(obj, "_doc4lint")
     doc = obj
     while not isinstance(doc, Tree.Document):
-        doc = getattr(doc, "parent")
+        if hasattr(doc, "_doc4lint"):
+            doc = getattr(doc, "_doc4lint")
+        else:
+            doc = getattr(doc, "parent")
         assert doc
     setattr(obj, "_doc4lint", doc)
     return doc
@@ -224,7 +225,7 @@ class StringCoercion(Linter):
             lambda from_type: not isinstance(
                 from_type,
                 (
-                    (Type.Any, Type.String, Type.File)  # pyre-ignore
+                    (Type.Any, Type.String, Type.File, Type.Directory)  # pyre-ignore
                     if isinstance(_parent_executable(obj), Tree.Task)
                     else (Type.Any, Type.String)
                 ),
@@ -235,32 +236,38 @@ class StringCoercion(Linter):
     def expr(self, obj: Expr.Base) -> Any:
         if isinstance(obj, Expr.Apply):
             # String function operands with non-String expression
-            if obj.function_name == "_add":
+            if obj.function_name in ("_add", "_interpolation_add"):
+                # TODO: should this apply to _interpolation_add, where coercion to String is
+                # "obviously" intended?
                 any_string = False
                 any_string_literal = False
-                all_string = True
+                non_string = None
                 for arg in obj.arguments:
                     if isinstance(arg.type, Type.String):
                         any_string = True
                         if isinstance(arg, Expr.String):
                             any_string_literal = True
-                    elif not isinstance(arg.type, Type.File):
-                        all_string = arg.type
-                if (
-                    any_string
-                    and all_string is not True
-                    # a literal string on one side or the other makes intent pretty clear
-                    and not any_string_literal
-                    # as does being inside an interpolation placeholder
-                    and not getattr(obj, "in_placeholder", False)
-                ):
-                    self.add(
-                        obj,
-                        "string concatenation (+) has {} argument; consider using interpolation".format(
-                            str(all_string)
-                        ),
-                        obj.pos,
-                    )
+                    elif not isinstance(arg.type, (Type.File, Type.Directory)):
+                        non_string = arg.type
+                if any_string and non_string:
+                    allowed = _find_doc(obj).effective_wdl_version in ("draft-2", "1.0")
+                    if not allowed:
+                        self.add(
+                            obj,
+                            "use interpolation instead of concatenating :String:"
+                            f" + :{non_string}: [deprecated in WDL >=1.1]",
+                            obj.pos,
+                        )
+                    elif not any_string_literal and obj.function_name != "_interpolation_add":
+                        # Prior to WDL 1.1, + could implicitly coerce a non-String argument to
+                        # concatenate with a String argument. Warn about this unless one side is a
+                        # a String literal or we're inside an interpolation (as those cases make
+                        # the intent clear)
+                        self.add(
+                            obj,
+                            f"consider interpolation instead of concatenating :String: + :{non_string}:",
+                            obj.pos,
+                        )
             else:
                 F = getattr(
                     StdLib.TaskOutputs(_find_doc(obj).effective_wdl_version), obj.function_name
@@ -277,7 +284,12 @@ class StringCoercion(Linter):
                             lambda from_type: not isinstance(
                                 from_type,
                                 (
-                                    (Type.Any, Type.String, Type.File)  # pyre-ignore
+                                    (
+                                        Type.Any,  # pyre-ignore
+                                        Type.String,
+                                        Type.File,
+                                        Type.Directory,
+                                    )
                                     if isinstance(_parent_executable(obj), Tree.Task)
                                     else (Type.Any, Type.String)
                                 ),
@@ -296,7 +308,7 @@ class StringCoercion(Linter):
             for elt in obj.items:
                 if isinstance(elt.type, Type.String):
                     any_string = True
-                elif not isinstance(elt.type, (Type.File, Type.Any)):
+                elif not isinstance(elt.type, (Type.File, Type.Directory, Type.Any)):
                     all_string = False
                 item_types.append(str(elt.type))
             if any_string and not all_string:
@@ -335,16 +347,19 @@ class FileCoercion(Linter):
     # exception: when rhs looks like a URI constant (typically a default reference database)
     def decl(self, obj: Tree.Decl) -> Any:
         super().decl(obj)
-        if (
-            obj.expr
-            and _compound_coercion(obj.type, obj.expr.type, (Type.File, Type.Directory))
-            and not (
+        if obj.expr and _compound_coercion(obj.type, obj.expr.type, (Type.File, Type.Directory)):
+            if (
                 isinstance(obj.expr, Expr.String)
                 and obj.expr.literal
                 and "://" in obj.expr.literal.value
-            )
-        ):
-            self.add(obj, "{} {} = :{}:".format(str(obj.type), obj.name, str(obj.expr.type)))
+            ):
+                self.add(
+                    obj,
+                    f'{obj.type} {obj.name} = "URI" may work with miniwdl, but for WDL portability,'
+                    " provide default URI in inputs JSON file",
+                )
+            else:
+                self.add(obj, "{} {} = :{}:".format(str(obj.type), obj.name, str(obj.expr.type)))
 
     def expr(self, obj: Expr.Base) -> Any:
         super().expr(obj)
@@ -466,25 +481,17 @@ class OptionalCoercion(Linter):
     def expr(self, obj: Expr.Base) -> Any:
         if isinstance(obj, Expr.Apply):
             if obj.function_name in ["_add", "_sub", "_mul", "_div", "_land", "_lor"]:
+                # excluded _interpolation_add, since interpolations expressly allow this
                 assert len(obj.arguments) == 2
                 arg0ty = obj.arguments[0].type
                 arg1ty = obj.arguments[1].type
-                if (arg0ty.optional or arg1ty.optional) and not (
-                    obj.function_name == "_add"
-                    and getattr(obj, "in_placeholder", False)
-                    and (
-                        isinstance(arg0ty, Type.String)
-                        and not arg0ty.optional
-                        or isinstance(arg1ty, Type.String)
-                        and not arg1ty.optional
-                    )
-                ):
-                    # exception for :String: + :T?: or vice-versa in string interpolations
+                if arg0ty.optional or arg1ty.optional:
                     self.add(
                         obj,
                         "infix operator has :{}: and :{}: operands".format(
                             str(arg0ty), str(arg1ty)
                         ),
+                        obj.pos,
                     )
             else:
                 F = getattr(
@@ -828,7 +835,7 @@ class UnnecessaryQuantifier(Linter):
                 and obj not in tw.inputs
                 and not (
                     isinstance(tw, Tree.Task)
-                    and isinstance(obj.type, Type.File)
+                    and isinstance(obj.type, (Type.File, Type.Directory))
                     and obj in tw.outputs
                 )
             ):
@@ -1079,8 +1086,7 @@ class UnboundDeclaration(Linter):
     # Unbound declaration outside of input{} section in WDL 1.0+
     def decl(self, obj: Tree.Decl) -> Any:
         if not obj.expr:
-            doc = _find_doc(obj)
-            if doc.wdl_version and doc.wdl_version != "draft-2":
+            if _find_doc(obj).effective_wdl_version != "draft-2":
                 exe = obj
                 while not isinstance(exe, (Tree.Task, Tree.Workflow)):
                     exe = getattr(exe, "parent")
@@ -1090,3 +1096,25 @@ class UnboundDeclaration(Linter):
                         obj,
                         f"{obj.type} {obj.name} should either be in the input section or bound to an expression",
                     )
+
+
+@a_linter
+class Deprecated(Linter):
+    def expr(self, obj: Expr.Base) -> Any:
+        if (
+            isinstance(obj, Expr.Placeholder)
+            and obj.options
+            and _find_doc(obj).effective_wdl_version not in ("draft-2", "1.0")
+        ):
+            self.add(
+                obj,
+                "use sep()/select_first()/if-then-else expressions instead of"
+                " sep/default/true/false placeholder options [WDL >= 1.1]",
+                obj.pos,
+            )
+        elif (
+            isinstance(obj, Expr.Struct)
+            and not obj.struct_type_name
+            and _find_doc(obj).effective_wdl_version not in ("draft-2", "1.0")
+        ):
+            self.add(obj, "replace 'object' with specific struct type [WDL >= 1.1]", obj.pos)
